@@ -15,6 +15,7 @@
 # limitations under the License.
 
 from glob import glob
+import itertools
 import os, subprocess, shutil, sys, signal
 from io import StringIO
 from ast import literal_eval
@@ -33,8 +34,10 @@ import time
 import multiprocessing
 import concurrent.futures as conc
 import re
+from run_unittests import get_fake_options, run_configure_inprocess
 
 from run_tests import get_backend_commands, get_backend_args_for_dir, Backend
+from run_tests import ensure_backend_detects_changes
 
 
 class BuildStep(Enum):
@@ -63,9 +66,6 @@ class DummyFuture(conc.Future):
     ask for the result. Used on platforms where sem_open() is not available:
     MSYS2, OpenBSD, etc: https://bugs.python.org/issue3770
     '''
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-
     def set_function(self, fn, *args, **kwargs):
         self.fn = fn
         self.fn_args = args
@@ -114,24 +114,11 @@ class AutoDeletedDir:
         return self.dir
 
     def __exit__(self, _type, value, traceback):
-        # On Windows, shutil.rmtree fails sometimes, because 'the directory is not empty'.
-        # Retrying fixes this.
-        # That's why we don't use tempfile.TemporaryDirectory, but wrap the deletion in the AutoDeletedDir class.
-        retries = 5
-        for i in range(0, retries):
-            try:
-                shutil.rmtree(self.dir)
-                return
-            # Sometimes we get: ValueError: I/O operation on closed file.
-            except ValueError:
-                return
-            # Deleting can raise OSError or PermissionError on Windows
-            # (most likely because of anti-virus locking the file)
-            except (OSError, PermissionError):
-                if i == retries - 1:
-                    mlog.warning('Could not delete temporary directory.')
-                    return
-                time.sleep(0.1 * (2**i))
+        # We don't use tempfile.TemporaryDirectory, but wrap the
+        # deletion in the AutoDeletedDir class because
+        # it fails on Windows due antivirus programs
+        # holding files open.
+        mesonlib.windows_proof_rmtree(self.dir)
 
 failing_logs = []
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
@@ -217,7 +204,6 @@ def validate_install(srcdir, installdir, compiler):
     # If this exists, the test does not install any other files
     noinst_file = 'usr/no-installed-files'
     expected = {}
-    found = {}
     ret_msg = ''
     # Generate list of expected files
     if os.path.exists(os.path.join(installdir, noinst_file)):
@@ -263,18 +249,6 @@ def log_text_file(logfile, testdir, stdo, stde):
         executor.shutdown()
         raise StopException()
 
-def run_configure_inprocess(commandlist):
-    old_stdout = sys.stdout
-    sys.stdout = mystdout = StringIO()
-    old_stderr = sys.stderr
-    sys.stderr = mystderr = StringIO()
-    try:
-        returncode = mesonmain.run(commandlist[0], commandlist[1:])
-    finally:
-        sys.stdout = old_stdout
-        sys.stderr = old_stderr
-    return returncode, mystdout.getvalue(), mystderr.getvalue()
-
 def run_test_inprocess(testdir):
     old_stdout = sys.stdout
     sys.stdout = mystdout = StringIO()
@@ -282,14 +256,19 @@ def run_test_inprocess(testdir):
     sys.stderr = mystderr = StringIO()
     old_cwd = os.getcwd()
     os.chdir(testdir)
+    test_log_fname = 'meson-logs/testlog.txt'
     try:
         returncode_test = mesontest.run(['--no-rebuild'])
+        if os.path.exists(test_log_fname):
+            test_log = open(test_log_fname, errors='ignore').read()
+        else:
+            test_log = ''
         returncode_benchmark = mesontest.run(['--no-rebuild', '--benchmark', '--logbase', 'benchmarklog'])
     finally:
         sys.stdout = old_stdout
         sys.stderr = old_stderr
         os.chdir(old_cwd)
-    return max(returncode_test, returncode_benchmark), mystdout.getvalue(), mystderr.getvalue()
+    return max(returncode_test, returncode_benchmark), mystdout.getvalue(), mystderr.getvalue(), test_log
 
 def parse_test_args(testdir):
     args = []
@@ -336,6 +315,10 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
         return TestResult('Test that should have failed succeeded', BuildStep.configure, stdo, stde, mesonlog, gen_time)
     if returncode != 0:
         return TestResult('Generating the build system failed.', BuildStep.configure, stdo, stde, mesonlog, gen_time)
+    # Touch the meson.build file to force a regenerate so we can test that
+    # regeneration works before a build is run.
+    ensure_backend_detects_changes(backend)
+    os.utime(os.path.join(testdir, 'meson.build'))
     # Build with subprocess
     dir_args = get_backend_args_for_dir(backend, test_build_dir)
     build_start = time.time()
@@ -350,16 +333,16 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
     if pc.returncode != 0:
         return TestResult('Compiling source code failed.', BuildStep.build, stdo, stde, mesonlog, gen_time, build_time)
     # Touch the meson.build file to force a regenerate so we can test that
-    # regeneration works. We need to sleep for 0.2s because Ninja tracks mtimes
-    # at a low resolution: https://github.com/ninja-build/ninja/issues/371
-    time.sleep(0.2)
+    # regeneration works after a build is complete.
+    ensure_backend_detects_changes(backend)
     os.utime(os.path.join(testdir, 'meson.build'))
     test_start = time.time()
     # Test in-process
-    (returncode, tstdo, tstde) = run_test_inprocess(test_build_dir)
+    (returncode, tstdo, tstde, test_log) = run_test_inprocess(test_build_dir)
     test_time = time.time() - test_start
     stdo += tstdo
     stde += tstde
+    mesonlog += test_log
     if should_fail == 'test':
         if returncode != 0:
             return TestResult('', BuildStep.test, stdo, stde, mesonlog, gen_time)
@@ -367,7 +350,7 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
     if returncode != 0:
         return TestResult('Running unit tests failed.', BuildStep.test, stdo, stde, mesonlog, gen_time, build_time, test_time)
     # Do installation, if the backend supports it
-    if len(install_commands) != 0:
+    if install_commands:
         env = os.environ.copy()
         env['DESTDIR'] = install_dir
         # Install with subprocess
@@ -383,12 +366,12 @@ def _run_test(testdir, test_build_dir, install_dir, extra_args, compiler, backen
     stde += e
     if pi.returncode != 0:
         return TestResult('Running clean failed.', BuildStep.clean, stdo, stde, mesonlog, gen_time, build_time, test_time)
-    if len(install_commands) == 0:
+    if not install_commands:
         return TestResult('', BuildStep.install, '', '', mesonlog, gen_time, build_time, test_time)
     return TestResult(validate_install(testdir, install_dir, compiler), BuildStep.validate, stdo, stde, mesonlog, gen_time, build_time, test_time)
 
 def gather_tests(testdir):
-    tests = [t.replace('\\', '/').split('/', 2)[2] for t in glob(os.path.join(testdir, '*'))]
+    tests = [t.replace('\\', '/').split('/', 2)[2] for t in glob(testdir + '/*')]
     testlist = [(int(t.split()[0]), t) for t in tests]
     testlist.sort()
     tests = [os.path.join(testdir, t[1]) for t in testlist]
@@ -405,33 +388,62 @@ def have_d_compiler():
         return True
     return False
 
+def have_objc_compiler():
+    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir='.')) as build_dir:
+        env = environment.Environment(None, build_dir, None, get_fake_options('/'), [])
+        objc_comp = env.detect_objc_compiler(False)
+        if not objc_comp:
+            return False
+        try:
+            objc_comp.sanity_check(env.get_scratch_dir(), env)
+        except:
+            return False
+        objcpp_comp = env.detect_objc_compiler(False)
+        if not objcpp_comp:
+            return False
+        try:
+            objcpp_comp.sanity_check(env.get_scratch_dir(), env)
+        except:
+            return False
+    return True
+
 def have_java():
     if shutil.which('javac') and shutil.which('java'):
         return True
     return False
 
 def detect_tests_to_run():
-    all_tests = []
-    all_tests.append(('common', gather_tests('test cases/common'), False))
-    all_tests.append(('failing-meson', gather_tests('test cases/failing'), False))
-    all_tests.append(('failing-build', gather_tests('test cases/failing build'), False))
-    all_tests.append(('failing-tests', gather_tests('test cases/failing tests'), False))
-    all_tests.append(('prebuilt', gather_tests('test cases/prebuilt'), False))
+    # Name, subdirectory, skip condition.
+    all_tests = [
+        ('common', 'common', False),
+        ('failing-meson', 'failing', False),
+        ('failing-build', 'failing build', False),
+        ('failing-tests', 'failing tests', False),
+        ('prebuilt', 'prebuilt', False),
 
-    all_tests.append(('platform-osx', gather_tests('test cases/osx'), False if mesonlib.is_osx() else True))
-    all_tests.append(('platform-windows', gather_tests('test cases/windows'), False if mesonlib.is_windows() or mesonlib.is_cygwin() else True))
-    all_tests.append(('platform-linux', gather_tests('test cases/linuxlike'), False if not (mesonlib.is_osx() or mesonlib.is_windows()) else True))
-    all_tests.append(('framework', gather_tests('test cases/frameworks'), False if not mesonlib.is_osx() and not mesonlib.is_windows() and not mesonlib.is_cygwin() else True))
-    all_tests.append(('java', gather_tests('test cases/java'), False if backend is Backend.ninja and not mesonlib.is_osx() and have_java() else True))
-    all_tests.append(('C#', gather_tests('test cases/csharp'), False if backend is Backend.ninja and shutil.which('mcs') else True))
-    all_tests.append(('vala', gather_tests('test cases/vala'), False if backend is Backend.ninja and shutil.which('valac') else True))
-    all_tests.append(('rust', gather_tests('test cases/rust'), False if backend is Backend.ninja and shutil.which('rustc') else True))
-    all_tests.append(('d', gather_tests('test cases/d'), False if backend is Backend.ninja and have_d_compiler() else True))
-    all_tests.append(('objective c', gather_tests('test cases/objc'), False if backend in (Backend.ninja, Backend.xcode) and not mesonlib.is_windows() else True))
-    all_tests.append(('fortran', gather_tests('test cases/fortran'), False if backend is Backend.ninja and shutil.which('gfortran') else True))
-    all_tests.append(('swift', gather_tests('test cases/swift'), False if backend in (Backend.ninja, Backend.xcode) and shutil.which('swiftc') else True))
-    all_tests.append(('python3', gather_tests('test cases/python3'), False if backend is Backend.ninja and shutil.which('python3') else True))
-    return all_tests
+        ('platform-osx', 'osx', not mesonlib.is_osx()),
+        ('platform-windows', 'windows', not mesonlib.is_windows() and not mesonlib.is_cygwin()),
+        ('platform-linux', 'linuxlike', mesonlib.is_osx() or mesonlib.is_windows()),
+
+        ('java', 'java', backend is not Backend.ninja or mesonlib.is_osx() or not have_java()),
+        ('C#', 'csharp', backend is not Backend.ninja or not shutil.which('mcs')),
+        ('vala', 'vala', backend is not Backend.ninja or not shutil.which('valac')),
+        ('rust', 'rust', backend is not Backend.ninja or not shutil.which('rustc')),
+        ('d', 'd', backend is not Backend.ninja or not have_d_compiler()),
+        ('objective c', 'objc', backend not in (Backend.ninja, Backend.xcode) or mesonlib.is_windows() or not have_objc_compiler()),
+        ('fortran', 'fortran', backend is not Backend.ninja or not shutil.which('gfortran')),
+        ('swift', 'swift', backend not in (Backend.ninja, Backend.xcode) or not shutil.which('swiftc')),
+        ('python3', 'python3', backend is not Backend.ninja),
+    ]
+    gathered_tests = [(name, gather_tests('test cases/' + subdir), skip) for name, subdir, skip in all_tests]
+    if mesonlib.is_windows():
+        # TODO: Set BOOST_ROOT in .appveyor.yml
+        gathered_tests += [('framework', ['test cases/frameworks/1 boost'], 'BOOST_ROOT' not in os.environ)]
+    elif mesonlib.is_osx() or mesonlib.is_cygwin():
+        gathered_tests += [('framework', gather_tests('test cases/frameworks'), True)]
+    else:
+        gathered_tests += [('framework', gather_tests('test cases/frameworks'), False)]
+    return gathered_tests
 
 def run_tests(all_tests, log_name_base, extra_args):
     global stop, executor, futures
@@ -642,4 +654,10 @@ if __name__ == '__main__':
         print('\nMesonlogs of failing tests\n')
         for l in failing_logs:
             print(l, '\n')
+    for name, dirs, skip in all_tests:
+        dirs = (os.path.basename(x) for x in dirs)
+        for k, g in itertools.groupby(dirs, key=lambda x: x.split()[0]):
+            tests = list(g)
+            if len(tests) != 1:
+                print('WARNING: The %s suite contains duplicate "%s" tests: "%s"' % (name, k, '", "'.join(tests)))
     sys.exit(failing_tests)
