@@ -16,11 +16,15 @@
 
 import os
 import sys
+import time
 import shutil
 import subprocess
+import tempfile
 import platform
 from mesonbuild import mesonlib
+from mesonbuild import mesonmain
 from mesonbuild.environment import detect_ninja
+from io import StringIO
 from enum import Enum
 from glob import glob
 
@@ -98,6 +102,13 @@ def get_backend_commands(backend, debug=False):
         raise AssertionError('Unknown backend: {!r}'.format(backend))
     return cmd, clean_cmd, test_cmd, install_cmd, uninstall_cmd
 
+def ensure_backend_detects_changes(backend):
+    # This is needed to increase the difference between build.ninja's
+    # timestamp and the timestamp of whatever you changed due to a Ninja
+    # bug: https://github.com/ninja-build/ninja/issues/371
+    if backend is Backend.ninja:
+        time.sleep(1)
+
 def get_fake_options(prefix):
     import argparse
     opts = argparse.Namespace()
@@ -106,14 +117,38 @@ def get_fake_options(prefix):
     opts.prefix = prefix
     return opts
 
+def should_run_linux_cross_tests():
+    return shutil.which('arm-linux-gnueabihf-gcc-6') and not platform.machine().startswith('arm')
+
+def run_configure_inprocess(commandlist):
+    old_stdout = sys.stdout
+    sys.stdout = mystdout = StringIO()
+    old_stderr = sys.stderr
+    sys.stderr = mystderr = StringIO()
+    try:
+        returncode = mesonmain.run(commandlist[0], commandlist[1:])
+    finally:
+        sys.stdout = old_stdout
+        sys.stderr = old_stderr
+    return returncode, mystdout.getvalue(), mystderr.getvalue()
+
 class FakeEnvironment(object):
     def __init__(self):
         self.cross_info = None
+        self.coredata = lambda: None
+        self.coredata.compilers = {}
 
     def is_cross_build(self):
         return False
 
 if __name__ == '__main__':
+    # Enable coverage early...
+    enable_coverage = '--cov' in sys.argv
+    if enable_coverage:
+        os.makedirs('.coverage', exist_ok=True)
+        sys.argv.remove('--cov')
+        import coverage
+        coverage.process_startup()
     returncode = 0
     # Iterate over list in reverse order to find the last --backend arg
     backend = Backend.ninja
@@ -143,18 +178,29 @@ if __name__ == '__main__':
         os.environ.pop('platform')
     # Run tests
     print('Running unittests.\n')
-    units = ['InternalTests', 'AllPlatformTests']
+    units = ['InternalTests', 'AllPlatformTests', 'FailureTests']
     if mesonlib.is_linux():
         units += ['LinuxlikeTests']
+        if should_run_linux_cross_tests():
+            units += ['LinuxArmCrossCompileTests']
     elif mesonlib.is_windows():
         units += ['WindowsTests']
     # Can't pass arguments to unit tests, so set the backend to use in the environment
     env = os.environ.copy()
     env['MESON_UNIT_TEST_BACKEND'] = backend.name
-    returncode += subprocess.call([sys.executable, 'run_unittests.py', '-v'] + units, env=env)
-    # Ubuntu packages do not have a binary without -6 suffix.
-    if shutil.which('arm-linux-gnueabihf-gcc-6') and not platform.machine().startswith('arm'):
-        print('Running cross compilation tests.\n')
-        returncode += subprocess.call([sys.executable, 'run_cross_test.py', 'cross/ubuntu-armhf.txt'])
-    returncode += subprocess.call([sys.executable, 'run_project_tests.py'] + sys.argv[1:])
+    with tempfile.TemporaryDirectory() as td:
+        # Enable coverage on all subsequent processes.
+        if enable_coverage:
+            with open(os.path.join(td, 'usercustomize.py'), 'w') as f:
+                f.write('import coverage\n'
+                        'coverage.process_startup()\n')
+            env['COVERAGE_PROCESS_START'] = '.coveragerc'
+            env['PYTHONPATH'] = os.pathsep.join([td] + env.get('PYTHONPATH', []))
+
+        returncode += subprocess.call([sys.executable, 'run_unittests.py', '-v'] + units, env=env)
+        # Ubuntu packages do not have a binary without -6 suffix.
+        if should_run_linux_cross_tests():
+            print('Running cross compilation tests.\n')
+            returncode += subprocess.call([sys.executable, 'run_cross_test.py', 'cross/ubuntu-armhf.txt'], env=env)
+        returncode += subprocess.call([sys.executable, 'run_project_tests.py'] + sys.argv[1:], env=env)
     sys.exit(returncode)
